@@ -26,6 +26,7 @@
 #include "interface.h"
 #include "byteorder.h"
 #include "global.h"
+#include "scsi_cmds.h"
 
 unsigned char *bufferTOC;
 unsigned char *cmd;
@@ -139,6 +140,10 @@ int allow_atapi(int new)
         if (new == old)
                 return (old);
 
+#if defined (__linux__) && defined (SG_EMULATED_HOST)
+	/* direct query with ioctl */
+	new = is_emulated_SCSI();
+#else
         jes_silent++;
 	if (new && jes_verbose) fprintf(stderr, "\nallow atapi: get all mode pages ...");
         if (new &&
@@ -146,7 +151,7 @@ int allow_atapi(int new)
                 new = FALSE;
         }
         jes_silent--;
-
+#endif
         is_atapi = new;
         return (old);
 }
@@ -250,7 +255,8 @@ int test_unit_ready(void)
 static int has_mode_page(int page, const char *pagename, int *lenp)
 {
 	u_char	mode[0x100];
-	int	len = 1;				/* Nach SCSI Norm */
+	int	hdlen;
+	int	len = 1;				/* obey SCSI */
 	int	try = 0;
 	struct	scsi_mode_page_header *mp;
 
@@ -261,7 +267,7 @@ again:
 
 	(void)test_unit_ready();
 	jes_silent++;
-/* Maxoptix bringt Aborted cmd 0x0B mit code 0x4E (overlapping cmds)*/
+	/* Maxoptix yields aborted cmd 0x0B with code 0x4E (overlapping cmds)*/
 
 	/*
 	 * The Matsushita CW-7502 will sometimes deliver a zeroed
@@ -281,16 +287,15 @@ again:
 
 	if (jes_verbose)
 		scsiprbytes("Mode Sense Data", mode, len - scsigetresid());
-	mp = (struct scsi_mode_page_header *)
-		(mode + sizeof(struct scsi_mode_header) +
-		((struct scsi_mode_header *)mode)->blockdesc_len);
+	hdlen = sizeof(struct scsi_mode_header) +
+		((struct scsi_mode_header *)mode)->blockdesc_len;
+	mp = (struct scsi_mode_page_header *)(mode + hdlen);
 	if (jes_verbose)
-		scsiprbytes("Mode Sense Data", (u_char *)mp, mp->p_len+2);
+		scsiprbytes("Mode Page  Data", (u_char *)mp, mp->p_len+2);
 
 	if (mp->p_len == 0) {
 		if (!scsi_compliant && try == 0) {
-			len = sizeof(struct scsi_mode_header) +
-			((struct scsi_mode_header *)mode)->blockdesc_len;
+			len = hdlen;
 			/*
 			 * add sizeof page header (page # + len byte)
 			 * (should normaly result in len == 14)
@@ -302,9 +307,19 @@ again:
 			try++;
 			goto again;
 		}
+		/* XXX if (!nowarn) */
 		errmsgno(EX_BAD,
 			"Warning: controller returns zero sized %s page.\n",
-								pagename);
+			pagename);
+	}
+	if (!scsi_compliant &&
+		(len < (mp->p_len + hdlen + 2))) {
+		len = mp->p_len + hdlen + 2;
+
+		/* XXX if (!nowarn) */
+		errmsgno(EX_BAD,
+			"Warning: controller returns wrong size for %s page.\n",
+			pagename);
 	}
 
 	if (lenp)
@@ -404,8 +419,10 @@ mmc_cap(unsigned char *modep)
                 ((struct scsi_mode_header *)mode)->blockdesc_len);
         /*
          * Do some heuristics against pre SCSI-3/mmc VU page 2A
+	 * We should test for a minimum p_len of 0x14, but some
+	 * buggy CD-ROM readers omit the write speed values.
          */
-        if (mp->p_len < 0x14)
+        if (mp->p_len < 0x10)
                 return (NULL);
 
         val = a_to_u_short(mp->max_read_speed);
@@ -570,8 +587,8 @@ void EnableCddaModeSelect (int fAudioMode)
   if (fAudioMode) {
     /* prepare to read audio cdda */
     mode [4] = density;  			/* cdda density */
-    mode [10] = (CD_FRAMESIZE_RAWER >> 8L);   /* block length "msb" */
-    mode [11] = (CD_FRAMESIZE_RAWER & 0xFF);
+    mode [10] = (cd_framesize >> 8L);   /* block length "msb" */
+    mode [11] = (cd_framesize & 0xFF);
   } else {
     /* prepare to read cds in the previous mode */
     mode [4] = orgmode4; /* 0x00; 			\* normal density */
@@ -1021,7 +1038,7 @@ unsigned ReadTocCdrSCSI ( TOC *toc )
 
 /* Read max. SectorBurst of cdda sectors to bufferCdda
    via standard SCSI-2 Read(10) command */
-void ReadStandard (unsigned char *p, long lSector, unsigned long SectorBurstVal )
+void ReadStandard (unsigned char *p, long lSector, unsigned long SectorBurstVal, int with_errorblockflags )
 {
   /* READ10, flags, block1 msb, block2, block3, block4 lsb, reserved, 
      transfer len msb, transfer len lsb, block addressing mode */
@@ -1029,7 +1046,7 @@ void ReadStandard (unsigned char *p, long lSector, unsigned long SectorBurstVal
 
         fillbytes((caddr_t)&scmd, sizeof(scmd), '\0');
         scmd.addr = (caddr_t)p;
-        scmd.size = SectorBurstVal*CD_FRAMESIZE_RAWER;
+        scmd.size = SectorBurstVal*cd_framesize;
         scmd.flags = SCG_RECV_DATA|SCG_DISRE_ENA;
         scmd.cdb_len = SC_G1_CDBLEN;
         scmd.sense_len = CCS_SENSE_LEN;
@@ -1049,7 +1066,7 @@ void ReadStandard (unsigned char *p, long lSector, unsigned long SectorBurstVal
 
 /* Read max. SectorBurst of cdda sectors to bufferCdda
    via vendor-specific ReadCdda(10) command */
-void ReadCdda10 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
+void ReadCdda10 (unsigned char *p, long lSector, unsigned long SectorBurstVal, int with_errorblockflags )
 {
   /* READ10, flags, block1 msb, block2, block3, block4 lsb, reserved, 
      transfer len msb, transfer len lsb, block addressing mode */
@@ -1057,7 +1074,7 @@ void ReadCdda10 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
 
         fillbytes((caddr_t)&scmd, sizeof(scmd), '\0');
         scmd.addr = (caddr_t)p;
-        scmd.size = SectorBurstVal*CD_FRAMESIZE_RAWER;
+        scmd.size = SectorBurstVal*cd_framesize;
         scmd.flags = SCG_RECV_DATA|SCG_DISRE_ENA;
         scmd.cdb_len = SC_G1_CDBLEN;
         scmd.sense_len = CCS_SENSE_LEN;
@@ -1077,13 +1094,13 @@ void ReadCdda10 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
 
 /* Read max. SectorBurst of cdda sectors to bufferCdda
    via vendor-specific ReadCdda(12) command */
-void ReadCdda12 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
+void ReadCdda12 (unsigned char *p, long lSector, unsigned long SectorBurstVal, int with_errorblockflags )
 {
   int count = 0;
 
         fillbytes((caddr_t)&scmd, sizeof(scmd), '\0');
         scmd.addr = (caddr_t)p;
-        scmd.size = SectorBurstVal*CD_FRAMESIZE_RAWER;
+        scmd.size = SectorBurstVal*cd_framesize;
         scmd.flags = SCG_RECV_DATA|SCG_DISRE_ENA;
         scmd.cdb_len = SC_G5_CDBLEN;
         scmd.sense_len = CCS_SENSE_LEN;
@@ -1104,13 +1121,13 @@ void ReadCdda12 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
 
 /* Read max. SectorBurst of cdda sectors to bufferCdda
    via MMC standard READ CD command */
-void ReadCddaMMC12 (unsigned char *p, long lSector, unsigned long SectorBurstVal )
+void ReadCddaMMC12 (unsigned char *p, long lSector, unsigned long SectorBurstVal, int with_errorblockflags )
 {
   int count = 0;
 
         fillbytes((caddr_t)&scmd, sizeof(scmd), '\0');
         scmd.addr = (caddr_t)p;
-        scmd.size = SectorBurstVal*CD_FRAMESIZE_RAWER;
+        scmd.size = SectorBurstVal*cd_framesize;
         scmd.flags = SCG_RECV_DATA|SCG_DISRE_ENA;
         scmd.cdb_len = SC_G5_CDBLEN;
         scmd.sense_len = CCS_SENSE_LEN;
@@ -1121,8 +1138,11 @@ void ReadCddaMMC12 (unsigned char *p, long lSector, unsigned long SectorBurstVal
         scmd.cdb.g5_cdb.res10 = 1; /* plus all subchannels RAW */
         g5_cdbaddr(&scmd.cdb.g5_cdb, lSector);
         g5x_cdblen(&scmd.cdb.g5_cdb, SectorBurstVal);
-	scmd.cdb.g5_cdb.count[3] = 1 << 4 | 3 << 5 ;    /* User data  + all headers */
-
+	scmd.cdb.g5_cdb.count[3] = with_errorblockflags
+		/* Sync + all headers + user data + EDC/ECC */
+		? 1 << 7 | 3 << 5 | 1 << 4 | 1 << 3 | 2 << 1
+		/* Sync + all headers + user data */
+		: 1 << 7 | 3 << 5 | 1 << 4 | 1 << 3;
 
         if (jes_verbose) fprintf(stderr, "\nReadMMC12 CDDA...");
   while (scsicmd("ReadCD MMC 12") < 0
